# Convert CSV spreadsheet of PDU structure definitions
# into source code for C++, C#, etc.
# $Id: emitPDUstructures.awk 304 2019-10-16 06:47:49Z geoff $
BEGIN {
    if (length (COMPILE_TARGET) > 0) {
        languageTarget = COMPILE_TARGET;
    } else {
        languageTarget = "c++"; # default
    }
    if (languageTarget == "python") {
        startCommentChar = "#";
        endCommentChar = "";
    } else {
        startCommentChar = "/*";
        endCommentChar = " */";
    }
    state=0;

    networkPrimitiveType["text"] = 1;   # admin type
    networkPrimitiveType["binary"] = 1; # admin type
    networkPrimitiveType["zero"] = 1;   # admin type
    networkPrimitiveType["textchar"] = 1;   # admin type
    networkPrimitiveType["char"] = 1;
    networkPrimitiveType["signed char"] = 1;
    networkPrimitiveType["unsigned char"] = 1;
    networkPrimitiveType["int8_t"] = 1;
    networkPrimitiveType["uint8_t"] = 1;
    networkPrimitiveType["int16_t"] = 2;
    networkPrimitiveType["uint16_t"] = 2;
    networkPrimitiveType["int32_t"] = 4;
    networkPrimitiveType["uint32_t"] = 4;
    networkPrimitiveType["int64_t"] = 8;
    networkPrimitiveType["uint64_t"] = 8;
    networkPrimitiveType["posix_nanoseconds"] = 8;   # admin type
    networkPrimitiveType["float"] = 4;
    networkPrimitiveType["double"] = 8;

    # Python struct module format strings
    pythonPackTable["text"] = "s";   # admin type
    pythonPackTable["binary"] = "B"; # admin type
    pythonPackTable["zero"] = "c";   # admin type
    pythonPackTable["textchar"] = "B"; # admin type
    pythonPackTable["char"] = "B";
    pythonPackTable["signed char"] = "b";
    pythonPackTable["unsigned char"] = "B";
    pythonPackTable["int8_t"] = "b";
    pythonPackTable["uint8_t"] = "B";
    pythonPackTable["int16_t"] = "h";
    pythonPackTable["uint16_t"] = "H";
    pythonPackTable["int32_t"] = "i";
    pythonPackTable["uint32_t"] = "I";
    pythonPackTable["int64_t"] = "q";
    pythonPackTable["uint64_t"] = "Q";
    pythonPackTable["posix_nanoseconds"] = "Q";
    pythonPackTable["float"] = "f";
    pythonPackTable["double"] = "d";

    # start as copy of networkPrimitiveType
    for (k in networkPrimitiveType) {
        sizeofTable[k] = networkPrimitiveType[k];
    }
#    sizeofTable["byte"] = 1;    # TODO: remove when C# sizeof works

    correctFieldName["PDU header"] = "pduHdr";
    correctFieldName["PDU Header"] = "pduHdr";
    correctFieldName["ProductKey header"] = "keyHdr";
    correctFieldName["productKeyHeader"] = "keyHdr";


    if (languageTarget == "csharp") {
        renameType["char"] = "byte";
        renameType["textchar"] = "byte";
        renameType["signed char"] = "sbyte";
        renameType["unsigned char"] = "byte";
        renameType["int8_t"] = "sbyte";
        renameType["uint8_t"] = "byte";
        renameType["int16_t"] = "short";
        renameType["uint16_t"] = "ushort";
        renameType["int32_t"] = "int";
        renameType["uint32_t"] = "uint";
        renameType["float"] = "float";
        renameType["int64_t"] = "long";
        renameType["uint64_t"] = "ulong";
        renameType["posix_nanoseconds"] = "ulong";
        renameType["double"] = "double";
        renameType["text"] = "byte";
        renameType["binary"] = "byte";
        renameType["zero"] = "byte";

        renameClassType["text"] = "string";
        renameClassType["binary"] = "string";
    } else if (languageTarget == "c++") {
        renameType["text"] = "char";
        renameType["textchar"] = "char";
        renameType["binary"] = "unsigned char";
        renameType["zero"] = "unsigned char";
        renameType["posix_nanoseconds"] = "uint64_t";

        renameClassType["text"] = "std::string";
        renameClassType["binary"] = "std::string";
    }

    fieldCount = 0;
    recordsOutput = 0;
    fieldNameToIndex["N/A"] = 0;
    arrayElementCount[0] = 0;
}

function stripSurroundingSpaces(s, _l, _i, _changed)
{
    _l = length(s);
    _changed = 0;
    while ((substr(s, _l, 1) == " ") && (_l > 1)) {
        _l -= 1;
        _changed += 1;
    }
    _i = 1;
    while ((_i < _l) && (substr(s, _i, 1) == " ")) {
        _i += 1;
        _changed += 1;
    }
    if (_changed != 0) {
        printf("STRIPPED %d: %s _i=%d _l=%d\n", _changed, s, _i, _l) > "/dev/stderr";
    }
    return (substr(s, _i, (_l - _i) + 1));
}

function dumpPDU_wireLayout()
{
    if (recordsOutput == 0) {
        printf("/* WARNING:  machine generated structures based on %s */\n",
            FILENAME);
        printf("/* target language: %s */\n", languageTarget);
        printf("\n");
        if (languageTarget == "csharp") {
            printf("/* $Id: emitPDUstructures.awk 304 2019-10-16 06:47:49Z geoff $ */\n");
            fieldTypePrefix = "public ";
            funcTypePrefix = "public ";
            typeSuffix = "";
            voidTypeName = "void ";
            printf("\nusing System.Runtime.InteropServices;\n");
            printf("\n");
            printf("public interface IAmPduWithKeyHdr {\n");
            printf("    PDUheader get_pduHdr();\n");
            printf("    void set_pduHdr(ref PDUheader arg);\n");
            printf("    ProductKeyHeader get_keyHdr();\n");
            printf("    void set_keyHdr(ref ProductKeyHeader arg);\n");
            printf("    uint get_checkSumCopy();\n");
            printf("    void set_checkSumCopy(uint arg);\n");
            printf("}\n");
            printf("\n");

            printf("public interface IAmPduWithKeyHdrNative {\n");
            printf("    PDUheader_Native PduHdr { get; set; }\n");
            printf("    ProductKeyHeader_Native KeyHdr { get; set; }\n");
            printf("}\n");
            printf("\n");
        } else {
            if (length(HEADER_FILENAME) == 0) {
                fileName = FILENAME;
            } else {
                fileName = HEADER_FILENAME;
            }
            l = length(fileName);
            foundDot = 0;
            for(i=l;i>=1;i-=1) {
                c = substr(fileName, i, 1);
                if ((foundDot == 0) && (c == ".")) {
                    foundDot = i - 1;
                }
                if (c == "/") break;
            }
            if (foundDot == 0) foundDot = l;
            defineName = "_" toupper(substr(fileName, i + 1, foundDot - i )) "_HPP_";
            
            printf("#ifndef %s\n", defineName);
            printf("#define %s \"%sId%s\"\n", defineName, "$", "$");
            printf("#include <utils/base/compiler_hints.h>\n");
            printf("#include <utils/time/time_point.hpp>\n");
            printf("#include <utils/logging/logging_api.hpp>\n");
            printf("#include <string>\n");
            printf("#include <stdint.h>\n");
            printf("#include <memory.h>\n");
            printf("#define _MIN_LEN(a,b) ((a <= b) ? (a) : (b))\n");
            printf("#ifndef _CRT_SECURE_NO_WARNINGS\n");
            printf("#define __CRT_WARN_NOT_ALREADY_DEFINED 1\n");
            printf("#define _CRT_SECURE_NO_WARNINGS 1 /* quiet Visual Studio C++ */\n");
            printf("#endif\n");
            printf("\ntypedef const struct StringInROM_struct *StringInROM;\n");
            printf("\n");
            fieldTypePrefix = "";
            funcTypePrefix = "inline ";
            voidTypeName = "void ";
            typeSuffix = "OME_ALWAYS_INLINE "
        }
    }
    extraQualifiers = "";
    # check to see if it contains special base classes
    for(i=0;i<fieldCount;i+=1) {
        t = fieldBaseType[i];
        if (t in renameType) {
            t = renameType[t];
        }
        n = fieldName[i];
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        if ((t == "ProductKeyHeader") && (n == "keyHdr")) {
            hasProductKeyHeader[PDUname] = n;
            if (languageTarget == "csharp") {
                extraQualifiers = " : IAmPduWithKeyHdr";
            }
#            break;
        }
        if (t == "TermHeaderStruct") {
            hasTermHeader[PDUname] = n;
        }
    }
    if (languageTarget == "csharp") {
        printf("[StructLayout(LayoutKind.Sequential,Pack=1)]\npublic unsafe ");
        printf("partial "); # for now
    }
    dependsUpon = "";
    sawDependentField = 0;
    printf("struct %s%s {\n", PDUname, extraQualifiers);
    for(i=0;i<fieldCount;i+=1) {
        t = fieldBaseType[i];
        if (t in renameType) {
            t = renameType[t];
        }
        n = fieldName[i];
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        if (arrayElementCount[i] != 0) {
            subscriptVal = substr(arrayElementCount[i], 1, 1);
            if ((subscriptVal < "0") || (subscriptVal > "9")) {
                sawDependentField = 1;
                dependendsUpon = n;
            }
        }
        if (sawDependentField == 0) {
            if (t in variableLengthStructs) {
#printf("%s is a variable length struct\n", n);
                sawDependentField = 1;
                dependsUpon = n;
            }
        }
        if (sawDependentField) {
            if (arrayElementCount[i] != 0) {
                printf("    /* %s %s implied field based on %s elements */\n", t, n, arrayElementCount[i]);
            } else {
                printf("    /* %s %s implied field */\n", t, n);
            }
            continue;
        }
        if (languageTarget == "csharp") {
            if (arrayElementCount[i] != 0) {
                t = "fixed " t;
            }
        }
        varName = n;
        if (arrayElementCount[i] != 0) {
            subscriptVal = substr(arrayElementCount[i], 1, 1);
            if ((subscriptVal >= "0") && (subscriptVal <= "9")) {
                varName = varName sprintf("[%u]", arrayElementCount[i]);
            } else { # should never ever happen
                varName = varName sprintf("[dep:%s]", arrayElementCount[i]);
            }
        }
        varName = varName ";"
        if (length(fieldComment[i]) > 0) {
            if (languageTarget == "csharp") {
                printf("    /// <summary>%s</summary>\n",
                       fieldComment[i]);
                printf("    %s", fieldTypePrefix); # indent
                printf("%-23s %-23s",
                    t, varName, fieldComment[i]);
            } else {
                printf("    %s", fieldTypePrefix); # indent
                printf("%-23s %-23s /*!< %s */",
                    t, varName, fieldComment[i]);
            }
        } else { # no comment
            printf("    %s", fieldTypePrefix); # indent
            printf("%-23s %s", t, varName);
        }
        printf("\n");
    }
    if ((sawDependentField == 1) && (languageTarget == "csharp")) {
        printf("    private fixed byte placeholder[%d]; /* force space to be allocated */\n",
               2048 - currentOffset);
    }
    # output member functions
    needTemporaryBuffer = 0;
    printf("\n");
    for(i=0;i<fieldCount;i+=1) {
        n = fieldName[i];
        if (substr(n, 1, 1) == "_") {   # skip fields beginning with "_"
            continue;
        }
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        t = fieldBaseType[i];
        declaredType = t;
        if (declaredType == "posix_nanoseconds") {
            needTemporaryBuffer = 1;
        }
        constModifier = ""; # assume not needed
        if (t in networkPrimitiveType) {
            # stuff that fits in registers
            fieldRef = "";
            retRef = "";
        } else {
            if (languageTarget == "csharp") {
                printf("    /* note: expensive copy */\n");
                fieldRef = "ref ";
                retRef = "";
            } else {
                fieldRef = "&";
                retRef = "&";
                constModifier = "const ";
            }
        }

        if (t in renameType) {
            t = renameType[t];
        }
        notArrayRef = fieldRef;
        elemRef = retRef;
        variableLengthField = ""; # assume false
        if (arrayElementCount[i] != 0) {
            deRefModifier = "";
            ptrModifier = "*";
            retRef = "";
            notArrayRef = "";
            if (languageTarget == "csharp") {
                constModifier = "";
            } else {
                constModifier = "const ";
            }
            s = substr(arrayElementCount[i], 1, 1);
            if ((s >= "0") && (s <= "9")) { # fixed-size field
                if (languageTarget == "csharp") {
                    subModifier = "[]";
                    ptrModifier = "[]";
                } else {
                    subModifier = sprintf("[%u]",
                        arrayElementCount[i]);
                }
            } else {
#printf("line %d variable length field %s, arrayElementCount=%s\n", NR, n, arrayElementCount[i]) >> "/dev/stderr";
                variableLengthField = arrayElementCount[i];
                subModifier = "[]";
                if (languageTarget == "csharp") {
                    ptrModifier = "[]";
                }
            }
            if (declaredType == "text") {
                subModifier = "[]";
                postArg = "";
            } else if (declaredType == "binary") {
                if (languageTarget != "csharp") {
                    subModifier = "[]";
                    postArg = ", size_t _len";
                } else {
                    subModifier = "[]";
                    postArg = ", uint _len";
                }
            }
        } else {
            ptrModifier = "";
            subModifier = "";
            postArg = "";
            deRefModifier = "*";
        }

        if (languageTarget == "csharp") {
            destName = "thisPtr->" n;
        } else {
            destName = n;
        }
        retDestName = destName;
        if (fieldDependsUpon[i] == "") { # simple variable
            cPtrStr = "";
            ptrStr = "";
        } else {
            dependsUpon = dependsUpon " " n;
            destName = "_resultPtr";
            cPtrStr = sprintf("        int _attrOffset = %s;\n",
                fieldOffset[i]);
            ptrStr = sprintf("        int _attrOffset = %s;\n",
                fieldOffset[i]);
            if (languageTarget != "csharp") {
                cPtrStr = cPtrStr sprintf("        const unsigned char *_basePtr = reinterpret_cast<const unsigned char *>(this);\n");
                cPtrStr = cPtrStr sprintf("        const %s *_resultPtr = reinterpret_cast<const %s *>(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
                ptrStr = ptrStr sprintf("        unsigned char *_basePtr = reinterpret_cast<unsigned char *>(this);\n");
                ptrStr = ptrStr sprintf("        %s *_resultPtr = reinterpret_cast<%s *>(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
            } else { # C# implementation
                cPtrStr = cPtrStr sprintf("        byte *_basePtr = (byte *)(thisPtr);\n");
                cPtrStr = cPtrStr sprintf("        %s *_resultPtr = (%s *)(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
                ptrStr = ptrStr sprintf("        byte *_basePtr = (byte *)(thisPtr);\n");
                ptrStr = ptrStr sprintf("        %s *_resultPtr = (%s *)(_basePtr + _attrOffset); /* %s */\n",
                t, t, n);
            }
        } # end else dependent on variable length field
        if (arrayElementCount[i] != 0) { # generate accessor for array
            printf("    %s", funcTypePrefix); # indent
            if (languageTarget != "csharp") {
                printf("%s%s %sgetElement_%s(const unsigned int _subscript) const %s{\n",
                    constModifier, t, elemRef, n, typeSuffix);
            } else {
                printf("%s %sgetElement_%s(uint _subscript) %s{\n",
                    t, elemRef, n, typeSuffix);
                    printf("      fixed(%s *thisPtr = &this) {\n", PDUname);
            }
            if (cPtrStr != "") {
                printf("%s", cPtrStr);
            }
            printf("        return (%s[_subscript]);\n", destName);
            if (languageTarget == "csharp") {
                printf("      }\n"); # close fixed scope
            }
            printf("    }\n");

            printf("    %s", funcTypePrefix); # indent
            if (languageTarget != "csharp") {
                printf("%ssetElement_%s(const unsigned int _subscript, const %s %sarg) %s{\n",
                    voidTypeName, n, t, fieldRef, typeSuffix);
            } else {
                printf("%ssetElement_%s(uint _subscript, %s arg) %s{\n",
                    voidTypeName, n, t, typeSuffix);
                    printf("      fixed(%s *thisPtr = &this) {\n", PDUname);
            }
            if (variableLengthField != "") {
                if (languageTarget != "csharp") {
                    printf("#ifndef NO_NULL_LENGTH_CHECKS\n");
                    printf("        if (get_%s() <= _subscript) { return; }\n", variableLengthField);
                    printf("#endif /* NO_NULL_LENGTH_CHECKS */\n");
                }
            }
            if (ptrStr != "") {
                printf("%s", ptrStr);
            }
            printf("        %s[_subscript] = arg;\n", destName);
            if (languageTarget == "csharp") {
                printf("      }\n"); # close fixed scope
            }
            printf("    }\n");
        } # end array



        printf("    %s", funcTypePrefix); # indent
        if (languageTarget == "csharp") {
            printf("%s%s %sget_%s() %s{\n",
                constModifier, t, ptrModifier, n, typeSuffix);
        } else {
            printf("%s%s %s%sget_%s() const %s{\n",
                constModifier, t, ptrModifier, retRef, n, typeSuffix);
        }
        scopeClose = "";
        resultScopeClose = "";
        addrRetOp = "";
        if (languageTarget == "csharp") {
            if (arrayElementCount[i] != 0) { # array copy
                printf("// note:  expensive copy to C# array\n");
                printf("        %s[] _resultAsArray = new %s[%s];\n",
                       t, t, arrayElementCount[i]);
                retDestName = "_resultAsArray";
                printf("      fixed(%s *thisPtr = &this) {\n", PDUname);
                printf("        uint _i = 0;\n");
                printf("        uint _l = %s;\n", arrayElementCount[i]);
                printf("        while (_i < _l) { /* memcpy */\n");
                if (declaredType == "text") { # stop copy on null
                    if (fieldDependsUpon[i] == "") { # simple variable
                        printf("            if (thisPtr->%s[_i] == 0) { // stop copying as null encountered\n", n);
                    } else {
                        printf("            if (getElement_%s(_i) == 0) { // stop copying as null encountered\n", n);
                    }
                    printf("                System.Array.Resize(ref %s, (int) _i); // truncate\n", retDestName);
                    printf("                break;\n");
                    printf("            }\n");
                } # end was of type text
                if (fieldDependsUpon[i] == "") { # simple variable
                    printf("            %s[_i] = thisPtr->%s[_i];\n", retDestName, n);
                } else {
                    printf("            %s[_i] = getElement_%s(_i);\n", retDestName, n);
                }
                printf("            _i += 1;\n");
                printf("        }\n");
                cPtrStr = "";
            } else { # not an array
                printf("      fixed(%s *thisPtr = &this) {\n", PDUname);
            }
            resultScopeClose = "      } // end result fixed scope\n";
        } else { # not csharp
            if (variableLengthField != "") {
                printf("#ifndef NO_NULL_LENGTH_CHECKS\n");
                if (declaredType == "text") {
                    printf("        if (get_%s() == 0) { return (\"NULL_%s\"); }\n", variableLengthField, variableLengthField);
                } else {
                    printf("        if (get_%s() == 0) { return (0); }\n", variableLengthField);
                }
                printf("#endif /* NO_NULL_LENGTH_CHECKS */\n");
            }
    }
if (0) {            
            cPtrStr = sprintf("        int _attrOffset = %s;\n",
                fieldOffset[i]);
            if (languageTarget != "csharp") {
                cPtrStr = cPtrStr sprintf("        const unsigned char *_basePtr = reinterpret_cast<const unsigned char *>(this);\n");
                cPtrStr = cPtrStr sprintf("        const %s *_resultPtr = reinterpret_cast<const %s *>(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
            } else { # C# implementation
                if (resultScopeClose == "") {
                    cPtrStr = cPtrStr sprintf("      fixed(%s *thisPtr = &this) {\n", PDUname);
                    scopeClose = "      } // end fixed scope\n";
                }
                cPtrStr = cPtrStr sprintf("        byte *_basePtr = (byte *)(thisPtr);\n");
                cPtrStr = cPtrStr sprintf("        %s *_resultPtr = (%s *)(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
            }
        } #end cPtrStr set block
        if (cPtrStr != "") {
            printf("%s", cPtrStr);
#            if (arrayElementCount[i] == 0) { # was not an array type, getElement_* was not used
                retDestName = sprintf("%s_resultPtr", deRefModifier);
#            }
        }
        destName = retDestName;

        printf("        return (%s%s);\n", addrRetOp, retDestName);
        printf("%s%s    }\n", resultScopeClose, scopeClose);

        printf("    %s", funcTypePrefix); # indent
        if (languageTarget == "csharp") {
            printf("%sset_%s(%s%s%s arg%s) %s{\n",
                voidTypeName, n, notArrayRef, t, subModifier, postArg, typeSuffix);
        } else {
            printf("%sset_%s(const %s %sarg%s%s) %s{\n",
                voidTypeName, n, t, notArrayRef, subModifier, postArg, typeSuffix);
        }
        if (languageTarget == "csharp") {
            printf("      fixed(%s *thisPtr = &this) {\n", PDUname);
        }
if (0) {
        if (fieldDependsUpon[i] == "") { # simple variable
            ptrStr = "";
        } else { # dependent variable
        if (0) {
            ptrStr = sprintf("        int _attrOffset = %s;\n",
                    fieldOffset[i]);
            if (languageTarget != "csharp") {
                ptrStr = ptrStr sprintf("        unsigned char *_basePtr = reinterpret_cast<unsigned char *>(this);\n");
                ptrStr = ptrStr sprintf("        %s *_resultPtr = reinterpret_cast<%s *>(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);
            } else {
                ptrStr = ptrStr sprintf("        byte *_basePtr = (byte *)(thisPtr);\n");
                ptrStr = ptrStr sprintf("        %s *_resultPtr = (%s *)(_basePtr + _attrOffset); /* %s */\n",
                    t, t, n);

            }
        }
            printf("%s", ptrStr);
        }
    }

        if (ptrStr != "") {
            printf("%s", ptrStr);
        }
        if ((arrayElementCount[i] == 0) ) {
            printf("        %s = arg;\n", destName);
        } else { # array
            c = substr(arrayElementCount[i], 1, 1);
            if ((c >= "0") && (c <= "9")) { # known size
                if (languageTarget != "csharp") {
                    field_size = sprintf("sizeof(%s)", n);
                } else {
                # TODO:  would really like to use sizeof()
#                   field_size = sprintf("sizeof(%s)", n); # Can't do in C#
                    field_size = arrayElementCount[i];
#                    destName = "thisPtr->" destName;
                }
            } else {
                # variable size
                if (languageTarget != "csharp") {
                    field_size = sprintf("(get_%s() * sizeof(%s))",
                        arrayElementCount[i], t);
                } else {
                # C# sizeof() is of type int, not uint
                    field_size = sprintf("((uint)(get_%s() * sizeof(%s)))",
                        arrayElementCount[i], t);
#                    field_size = sprintf("(uint) (get_%s() * %d)",
#                        arrayElementCount[i], sizeofTable[t]);
                }
            }
            if (languageTarget != "csharp") {
                if (declaredType == "text") {
                    printf("        strncpy(%s, arg, %s);\n",
                        destName, field_size);
                } else if (declaredType == "binary") {

                    printf("        memcpy(%s, arg, _MIN_LEN(_len,%s));\n",
                        destName, field_size);
                } else {
                    printf("        memcpy(%s, arg, %s);\n",
                        destName, field_size);
                }
            } else { # do C-sharp
                    printf("        uint _i = 0;\n");
                    printf("        uint _l = %s;\n", field_size);
                    printf("        while (_i < _l) { /* strncpy */\n", field_size);
                    printf("            if (_i == arg.Length) break;\n");
                    if (declaredType == "text") {
                        printf("            if (arg[_i] == 0) break;\n");
                    }
                    if (fieldDependsUpon[i] == "") { # simple variable
                        printf("            thisPtr->%s[_i] = arg[_i];\n", n);
                    } else {
                        printf("            setElement_%s(_i, arg[_i]);\n", n);
                    }
                    printf("            _i += 1;\n");
                    printf("        }\n");
                    if (declaredType == "text") {
                        printf("        while (_i < _l) { /* zero fill any remainder */\n", field_size);
                        if (fieldDependsUpon[i] == "") { # simple variable
                            printf("            thisPtr->%s[_i] = 0;\n", n);
                        } else {
                            printf("            setElement_%s(_i,0);\n", n);
                        }
                        printf("            _i += 1;\n");
                        printf("        }\n");
                    }
                }
                if (0) {
                    printf("        uint _i = 0;\n");
                    printf("        uint _l = %s;\n", arrayElementCount[i]);
                    printf("        while (_i < _l) { /* memcpy */\n", field_size);
                    printf("            %s[_i] = arg[_i];\n", destName);
                    printf("            _i += 1;\n");
                    printf("        }\n");
                }
            }
        printf("%s%s    }\n", resultScopeClose, scopeClose);

    } # end emit member accessor functions
    printf("\n");
    if (languageTarget == "csharp") {
        if (dependsUpon == "") {
            staticLabel = "static ";
        } else {
            staticLabel = "";
        }
        printf("    public %suint get_expectedSize() {\n", staticLabel);
        printf("        return ((uint) (%d%s));\n", currentOffset, dependentOffset);
        print ("    }\n");
    } else { // # C++
        if (sawDependentField == 0) {
            constExpr = "CONSTEXPR "; # needs trailing space
        } else {
            constExpr = "";
        }
        printf("    %sunsigned int get_expectedSize() const %s{\n", constExpr, typeSuffix);
        printf("        return (%d%s);\n", currentOffset, dependentOffset);
        print ("    }\n");
        if (sawDependentField != 0) {
            printf("    %s & operator=(const %s &arg) { /* assignment operator */\n", PDUname, PDUname);
            printf("        uint32_t len = arg.get_expectedSize();\n");
            printf("        memcpy(this, &arg, len);\n");
            printf("        return (*this);\n");
            print ("    }\n");
        }
    }

    if ((languageTarget == "csharp") && (sawDependentField == 0)) {
        # emit constructor from buffer
        if (PDUname == "PacketHeader") {
            lenFieldName = "srcPtr->headerLength";
        } else if (PDUname == "PDUheader") {
            lenFieldName = "srcPtr->pduLength";
        } else if (PDUname == "ProductKeyHeader") {
            lenFieldName = "sizeof(ProductKeyHeader)";
        } else {
            lenFieldName = "srcPtr->pduHdr.pduLength";
        }
        printf("\n");
        printf("public static %s get_struct_from_buffer(byte[] bfr, uint offset)\n",
               PDUname);
        printf("{\n");
#        printf("  %s result;\n", PDUname);
        printf("    GCHandle handle = GCHandle.Alloc(bfr, GCHandleType.Pinned);\n");
        printf("    System.Object obj = Marshal.PtrToStructure(System.IntPtr.Add(handle.AddrOfPinnedObject(), (int) offset), typeof(%s));\n", PDUname);
#        printf("  fixed(byte *bfrPtr = bfr) {\n");
#        printf("    %s *srcPtr = (%s *) (bfrPtr + offset);\n", PDUname, PDUname);
#        printf("    byte[] copiedBytes = new byte[%s];\n", lenFieldName);
#        printf("    System.Buffer.BlockCopy(bfr, offset, copiedBytes, 0, %s);\n", lenFieldName);
#        printf("    System.Object obj = copiedBytes;\n");
#        printf("    System.IntPtr iPtr = new System.IntPtr(copiedBytes);\n");
#        printf("    System.Object obj = Marshal.PtrToStructure(System.IntPtr.Add(iPtr, offset), typeof(PacketHeader));\n");
#        printf("    System.Object obj = Marshal.PtrToStructure(iPtr, typeof(PacketHeader));\n");
        printf("    %s result = (%s) obj;\n", PDUname, PDUname);
        printf("    handle.Free();\n");
        printf("    return (result);\n");
#        printf("  }\n");
        printf("}\n");
    }
    if (languageTarget == "csharp") {
        printf("\npublic string toString() {\n");
        printf("    System.Text.StringBuilder result = new System.Text.StringBuilder();\n");
        printf("    System.Text.ASCIIEncoding encoder = new System.Text.ASCIIEncoding();\n");
        printf("    result.Append(\"[%s={\");\n", PDUname);
       for(i=0;i<fieldCount;i+=1) {
            n = fieldName[i];
            if (substr(n, 1, 1) == "_") {   # skip fields beginning with "_"
                continue;
            }
            if (n in correctFieldName) {
                n = correctFieldName[n];
            }
            t = fieldBaseType[i];
            declaredType = t;
            doLoop = 0;
            if (arrayElementCount[i] != 0) {
                if (declaredType != "text") { # not a string
                    doLoop = 1;
                    subscriptVal = substr(arrayElementCount[i], 1, 1);
                    if ((subscriptVal >= "0") && (subscriptVal <= "9")) {
                        doLoopVar = arrayElementCount[i];
                    } else {
                        doLoopVar = "this.get_" arrayElementCount[i] "()";
                    }
                }
            }
            if (doLoop != 0) {
                printf("    for(uint _i = 0; _i < %s; _i+=1) {\n    ",
                    doLoopVar);
            }
            printf("    result.Append(\" %s", n);
            if (doLoop != 0) {
                printf("[\");\n");     
                printf("        result.Append(_i);\n");
                printf("        result.Append(\"]");
            }
            printf("=");
            if ((declaredType == "text") || (declaredType == "textchar")) {
                printf("\\\"");
            }
            printf("\");\n");
            if (doLoop != 0) { # add more indentation
                printf("    ");
            }
            printf("    result.Append(");
            if (declaredType == "text") {
                printf("encoder.GetString(");
            } else if (declaredType == "textchar") {
                printf("(char) ");
            }
            if (doLoop != 0) {
                printf("this.getElement_%s(_i)", n);
            } else {
                printf("this.get_%s()", n);
            }
            if (t in networkPrimitiveType) {
                if (declaredType == "text") {
                    printf(")"); # close encoder.GetString() call
                }
            } else {
                printf(".toString()");
            }
            printf("); /* output value */\n");
            if ((declaredType == "text") || (declaredType == "textchar")) {
                printf("    result.Append(\"\\\"\");\n");
            } else if (declaredType == "posix_nanoseconds") {
#                printf(").formatIntoString(bfr, sizeof(bfr), 0, NanosecondTimePoint::OUTPUT_DATE|NanosecondTimePoint::OUTPUT_TIME, \"/:.@\")");
            }
            if (doLoop != 0) {
                printf("    }\n");
            }
        }
        printf("    result.Append(\" }]\");\n");
        printf("    return (result.ToString());\n");
        printf("}\n");
    }

    printf("\n");
    printf("}; /* end struct %s */\n", PDUname);
    printf("/* expected length = %d%s */\n", currentOffset, dependentOffset);
    printf("\n");
    if (languageTarget == "c++") {
        printf("template <typename STREAMTYPE> inline STREAMTYPE & operator<<(STREAMTYPE &s, const %s &pdu) {\n", PDUname);
        if (needTemporaryBuffer == 1) {
            printf("    char bfr[64];\n\n");
        }
        printf("    s << (StringInROM) \"[%s={\";\n", PDUname);
        for(i=0;i<fieldCount;i+=1) {
            n = fieldName[i];
            if (substr(n, 1, 1) == "_") {   # skip fields beginning with "_"
                continue;
            }
            if (n in correctFieldName) {
                n = correctFieldName[n];
            }
            t = fieldBaseType[i];
            declaredType = t;
            doLoop = 0;
            if (arrayElementCount[i] != 0) {
                if (fieldBaseType[i] != "text") { # not a string
                    doLoop = 1;
                    # packet can't be more than 64K in the first place...
                    loopVarType = "uint_fast16_t";
                    subscriptVal = substr(arrayElementCount[i], 1, 1);
                    if ((subscriptVal >= "0") && (subscriptVal <= "9")) {
                        doLoopVar = arrayElementCount[i];
                        loopVarVal = 1 * doLoopVar;
                        if ((loopVarVal < 255) && (loopVarVal > 0)) {
                            loopVarType = "uint_fast8_t";
                        }
                    } else {
                        doLoopVar = "pdu.get_" arrayElementCount[i] "()";
                    }
                }
            }
            if (doLoop != 0) {
                printf("    for(%s _i = 0; _i < %s; ++_i) {\n    ",
                    loopVarType, doLoopVar);
            }

            printf("    s << (StringInROM) \" %s", n);
            if (doLoop != 0) {
                printf("[\" << (uint32_t) _i << (StringInROM) \"]");
            }
            printf("=");
            if ((declaredType == "text") || (declaredType == "textchar")) {
                printf("\\\"");
            }
            printf("\" << ");
            if (declaredType == "text") {
                printf("AS_ESCAPED_TEXT_BUFFER(");
            }
            if (declaredType == "textchar") {
                printf("(char) ");
            } else if (declaredType == "binary") {
                printf("(uint32_t) ");
            } else if (declaredType == "unsigned char") {
                printf("(uint32_t) ");
            } else if (declaredType == "uint8_t") {
                printf("(uint32_t) ");
            } else if (declaredType == "signed char") {
                printf("(int32_t) ");
            } else if (declaredType == "int8_t") {
                printf("(int32_t) ");
            } else if (declaredType == "posix_nanoseconds") {
                printf("NanosecondTimePoint(0, ");
            }
            if (doLoop != 0) {
                printf("pdu.getElement_%s(_i)", n);
            } else {
                printf("pdu.get_%s()", n);
            }
            if (declaredType == "text") {
                printf(")"); # close AS_ESCAPED_TEXT_BUFFER()
            }
            if ((declaredType == "text") || (declaredType == "textchar")) {
                printf(" << (StringInROM) \"\\\"\"");
            } else if (declaredType == "posix_nanoseconds") {
                printf(").formatIntoString(bfr, sizeof(bfr), 0, NanosecondTimePoint::OUTPUT_DATE|NanosecondTimePoint::OUTPUT_TIME, \"/:.@\")");
            }
            printf(";\n");
            if (doLoop != 0) {
                printf("    }\n");
            }
        }
        printf("    s << (StringInROM) \" }]\";\n");
        printf("    return (s);\n");
        printf("}\n");
    } # end output C++ print function
    printf("\n");
    recordsOutput += 1;
}

function dumpPythonPDU_layout()
{
    if (recordsOutput == 0) {
        printf("# WARNING:  machine generated structures based on %s\n",
            FILENAME);
        printf("# target language: %s\n", languageTarget);
        printf("\n");
        printf("# $Id: emitPDUstructures.awk 304 2019-10-16 06:47:49Z geoff $\n");
        printf("from struct import *\n\n");
    }
    printf("class %s:\n", PDUname);
    dependsUpon = "";
    sawDependentField = 0;
    packString = "";
    for(i=0;i<fieldCount;i+=1) {
        t = fieldBaseType[i];
        declaredType = t;
        if (t in renameType) {
            t = renameType[t];
        }
        n = fieldName[i];
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        if (arrayElementCount[i] != 0) {
            subscriptVal = substr(arrayElementCount[i], 1, 1);
            if ((subscriptVal < "0") || (subscriptVal > "9")) {
                sawDependentField = 1;
                dependendsUpon = n;
            }
        }
        if (sawDependentField) {
            if (arrayElementCount[i] != 0) {
                printf("# %s %s implied field based on %s elements\n", t, n, arrayElementCount[i]);
            } else {
                printf("# %s %s implied field\n", t, n);
            }
            continue;
        }
        packElement = pythonPackTable[t];
        if (arrayElementCount[i] != 0) {
            subscriptVal = substr(arrayElementCount[i], 1, 1);
            if ((subscriptVal >= "0") && (subscriptVal <= "9")) {
                packElement = sprintf("%u%s", arrayElementCount[i], packElement);
            } else { # should never ever happen
                varName = varName sprintf("[dep:%s]", arrayElementCount[i]);
            }
        }
        packString = packString packElement;
        printf("    def get_%s(self):\n", n);
        printf("        return struct.unpack_from(\"%s\", self.bfr, %u)\n",
               packElement, fieldOffset[i]);
        printf("\n");
        printf("    def set_%s(self, arg):\n", n);
        printf("        struct.pack_into(\"%s\", self.bfr, %u, arg)\n",
               packElement, fieldOffset[i]);
        printf("\n");
    }
    pythonPackTable[PDUname] = packString;
    printf("    %s_packFormat = \"%s\"\n", PDUname, packString);
    printf("\n");
    recordsOutput += 1;
}

function dumpPDU_nativeLayout()
{
    extraQualifiers = "";
    if (languageTarget == "csharp") {
        # check to see if it contains special base classes
        for(i=0;i<fieldCount;i+=1) {
            t = fieldBaseType[i];
            if (t in renameType) {
                t = renameType[t];
            }
            n = fieldName[i];
            if (n in correctFieldName) {
                n = correctFieldName[n];
            }
            if ((t == "ProductKeyHeader") && (n == "keyHdr")) {
                extraQualifiers = " : IAmPduWithKeyHdrNative";
                break;
            }
        }
        typePrefix = "public ";
        printf("public ");
    } else {
        typePrefix = "";
    }
    dependsUpon = "";
    printf("class %s_Native%s {\n", PDUname, extraQualifiers);
    for(i=0;i<fieldCount;i+=1) {
        n = fieldName[i];
        if (substr(n, 1, 1) == "_") { # skip any _pad fields
            continue;
        }
        t = fieldBaseType[i];
        # convert to high-level abstraction
        if (t in renameClassType) {
            t = renameClassType[t];
        }
        constModifier = ""; # assume not needed
        nativeQualifier = "";
if (1) { # DELETE
        if (t in networkPrimitiveType) {
            # stuff that fits in registers
            fieldRef = "";
        } else {
            nativeQualifier = "_Native";
            if (languageTarget == "csharp") {
                fieldRef = "ref ";
            } else {
                fieldRef = "&";
                constModifier = "const ";
            }
        }
} # END DELETE
        if (t in renameType) { # convert to language-specific
            t = renameType[t];
        }
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        printf("    %s", typePrefix); # indent
        if (languageTarget == "csharp") {
            if (arrayElementCount[i] != 0) {
                if (t != "string") { # not a string
                    nativeQualifier = "[]";
                } else {
                    nativeQualifier = "";
                }
            }
        }
        printf("%-23s %s", t nativeQualifier, n);
        if (languageTarget == "c++") {
            if (arrayElementCount[i] != 0) {
                if (fieldBaseType[i] != "text") { # not a string
                    printf("[%u]", arrayElementCount[i]);
                }
            }
        }
        printf(";");
        if (length(fieldComment[i]) > 0) {
            printf("\t/* %s */", fieldComment[i]);
        }
        printf("\n");
    }
    # output member functions
    printf("\n");
    if (extraQualifiers != "") {
        printf("    public PDUheader_Native PduHdr {\n");
        printf("        get { return pduHdr; }\n");
        printf("        set { pduHdr = value; }\n");
        printf("    }\n");
        printf("    public ProductKeyHeader_Native KeyHdr {\n");
        printf("        get { return keyHdr; }\n");
        printf("        set { keyHdr = value; }\n");
        printf("    }\n");
    }
    constModifier = ""; # assume not needed
    if (languageTarget == "csharp") {
        printf("    /* note: expensive copy */\n");
        printf("    public ");
        fieldRef = "";
        constModifier = ""; # "ref ";
        callByRef = ""; # "ref ";
    } else {
        printf("    ");
        fieldRef = "&";
        constModifier = "const ";
        callByRef = "";
    }
    printf("void copyFrom_struct(%s%s %swireFormat) {\n",
        constModifier, PDUname, fieldRef);
    if (languageTarget == "csharp") {
        printf("       System.Text.ASCIIEncoding encoder = new System.Text.ASCIIEncoding();\n");
    }
#   printf("      do {\n");
    for(i=0;i<fieldCount;i+=1) {
        n = fieldName[i];
        if (substr(n, 1, 1) == "_") { # skip any _ fields
            continue;
        }
        if (n in correctFieldName) {
            n = correctFieldName[n];
        }
        t = fieldBaseType[i];
        # convert to high-level abstraction
        if (t in renameClassType) {
            t = renameClassType[t];
        }
        if (t in networkPrimitiveType) {
            # stuff that fits in registers
            printf("        %s = wireFormat.get_%s();\n", n, n);
        } else {
            if (arrayElementCount[i] != 0) {
                if (languageTarget == "csharp") {
                    if (t == "string") {
            printf("        %s = encoder.GetString(wireFormat.get_%s());\n", n, n);
                    }
                }
            } else {
                printf("        %s.copyFrom_struct(%s wireFormat.get_%s());\n", n, callByRef, n);
            }
        }
    }
#   printf("      } (while pduHdr.checkSum != checkSumCopy);\n");
    printf("    }\n");
    printf("\n");
    printf("    %s_Native(%s%s %swireFormat) {\n",
        PDUname, constModifier, PDUname, fieldRef);
    printf("        copyFrom_struct(wireFormat);\n", PDUname);
    printf("    }\n");
    printf("}; /* end class %s_Native */\n", PDUname);
    printf("\n");
}

function dumpEnumTables()
{
    for (group in enumGroupCount) {
        count = enumGroupCount[group];
        if (languageTarget == "csharp") {
            printf("public ");
        }
        if (languageTarget == "python") {
            printf("class %s:\n", group);
            elemSep = "";
        } else {
            printf("enum %s {\n", group);
            elemSep = ",";
        }
        for(i=1;i<=count;i+=1) {
            if (i != count) {
                name = enumTable[group,i] elemSep;
            } else {
                name = enumTable[group,i];
            }

            if (length(enumTableComment[group,i]) > 0) {
                if (languageTarget == "csharp") {
                    printf("    /// <summary>%s</summary>\n",
                        enumTableComment[group,i])
                    printf("    %-47s\n", name);
                } else {
                    printf("    %-47s /*!< %s */\n",
                        name, enumTableComment[group,i])
                }
            } else { # no comment
                printf("    %s\n", name);
            }
        }
        if (languageTarget != "python") {
            printf("};\n");
        }
        printf("\n");
    }
}

function dumpEnumMacros()
{
    for (group in enumGroupCount) {
        if (group != "enPDUtype") {
            continue;
        }
        printf("#ifndef EXPECTED_PDU_TYPE\n");
        printf("/*! \\brief Return the expected value for pduHdr.pduType to be\n");
        printf(" * set into the indicated structure.\n");
        printf(" *\n");
        printf(" * \\param x denotes the PDU structure to be passed to typeid().\n");
        printf(" *\n");
        printf(" * Note: the seemingly ridiculously long sequence of ? operators can be\n");
        printf(" * evaluated at compile-time and results in a constant value.\n");
        printf(" */\n");

        printf("#define EXPECTED_PDU_TYPE(x) ( \\\n");
        closeText = "    enPDUtype_None )"
        count = enumGroupCount[group];
        for(i=1;i<=count;i+=1) {
            name = enumTable[group,i];
            j = index(name, "=");
            if (j != 0) {
                name = substr(name, 1, j - 1);
            }
            if (name == "enPDUtype_None") {
                continue;
            }
            # safety check on name
            if (substr(name, 1, 10) == "enPDUtype_") {
                pduType = substr(name, 11);
                pduName = pduType "PDU";
                printf("    ( typeid(x) == typeid(%s) ? %s : \\\n",
                       pduName, name);
                closeText = closeText ")";
            }
        }
        printf("%s\n", closeText);
        printf("#endif /* EXPECTED_PDU_TYPE */\n\n");
    }
    if (length(hasProductKeyHeader) > 0) {
        printf("#ifndef PDU_HAS_PRODUCT_KEY_HEADER\n");
        printf("/*! \\brief Return a Boolean value indicating if a PDU type has\n");
        printf(" * a ProductKeyHeader structure.\n");
        printf(" *\n");
        printf(" * \\param pduType is the integer enPDUtype enum value to be tested.\n");
        printf(" *\n");
        printf(" * \\retval false indicates the PDU does not contain a ProductKeyHeader.\n");
        printf(" * \\retval true indicates the PDU does contain a ProductKeyHeader.\n");
        printf(" */\n");
        printf("#define PDU_HAS_PRODUCT_KEY_HEADER(pduType) (( \\\n");
        for (pduName in hasProductKeyHeader) {
            pduType = substr(pduName, 1, length(pduName) - 3);
            printf("    ((pduType) == enPDUtype_%s) || \\\n", pduType);
        }
        printf("    0) ? true : false)\n");
        printf("#endif /* PDU_HAS_PRODUCT_KEY_HEADER */\n\n");
    }
    # TODO: PDU_HAS_TERM_HEADER_STRUCT(pduType)
        printf("#ifndef PDU_HAS_TERM_HEADER_STRUCT\n");
        printf("#define PDU_HAS_TERM_HEADER_STRUCT(pduType) ( false )\n");
        printf("#endif /* PDU_HAS_TERM_HEADER_STRUCT */\n\n");

    printf("/*! \\brief Outputs an object onto a stream.\n");
    printf(" * \\param onStream names the stream-like object that supports\n");
    printf(" *  operator<<() operations.\n");
    printf(" * \\param asType names the expected type of the object to be output.\n");
    printf(" * \\param p is a pointer to the object.\n");
    printf(" */\n");
    printf("#define PRINT_PDU_AS_TYPE(onStream, asType, p) onStream << *((const asType *)(p))\n\n");

    printf("// forward declaration\n");
    printf("template <typename STREAMTYPE> int formatPacketContentsOnStream(STREAMTYPE &outStream,\n");
    printf("    const PacketHeader *pkt, const PDUheader *pduHdr);\n\n");


    printf("/*! \\brief Templatized output routine that formats an optional packet header\n");
    printf(" * and PDU for output onto a stream-like object.\n");
    printf(" *\n");
    printf(" * \\param outStream is the stream-like object to which the formatted text\n");
    printf(" *  should be added via the \"<<\" operator.\n");
    printf(" * \\param pkt is a possibly null pointer to a PacketHeader.  If not null,\n");
    printf(" * the contents are formatted onto the \\a outStream.\n");
    printf(" * \\param pduHdr is a possibly null pointer to a PDUheader.  If not null,\n");
    printf(" * the \\a pduType field is examined to obtain the intended logical PDU type\n");
    printf(" * and the corresponding content is output to the \\a outStream.\n");
    printf(" * If the implied logical type is unknown, only the contents of the header will\n");
    printf(" * be output.\n");
    printf(" *\n");
    printf(" * \\retval -1 indicates the PDU type was not recognized.\n");
    printf(" * \\retval 0 indicates the PDU type was handled.\n");

    printf(" */\n");
    printf("template <typename STREAMTYPE> inline int formatPDUonStream(STREAMTYPE &outStream,\n");
    printf("    const PacketHeader *pkt, const PDUheader *pduHdr)\n");
    printf("{\n");
    printf("    if (pkt != nullptr) {\n");
    printf("        outStream << *pkt;\n");
    printf("        if (pduHdr != nullptr) {\n");
    printf("            outStream << \" \";\n");
    printf("        }\n");
    printf("    }\n");
    printf("    if (OME_EXPECT_TRUE(pduHdr != nullptr)) {\n");
    printf("        switch (pduHdr->pduType) {\n"); 

    for (group in enumGroupCount) {
        if (group != "enPDUtype") {
            continue;
        }
        count = enumGroupCount[group];
        for(i=1;i<=count;i+=1) {
            name = enumTable[group,i];
            j = index(name, "=");
            if (j != 0) {
                name = substr(name, 1, j - 1);
            }
            if (name == "enPDUtype_None") {
                continue;
            }
            # safety check on name
            if (substr(name, 1, 10) == "enPDUtype_") {
                pduType = substr(name, 11);
                pduName = pduType "PDU";
                printf("        case %s:\n", name);
                if (pduType == "Respond") {
                    printf("          {\n");
                    printf("            // print encapsulating header\n");
                }
                printf("            PRINT_PDU_AS_TYPE(outStream,%s,pduHdr);\n",
                       pduName);
                if (pduType == "Respond") {
                    printf("            // recursively handle encapsulated PDU\n");
                    printf("            const unsigned char *nextPDU = reinterpret_cast<const unsigned char *>(pduHdr) + pduHdr->pduHeaderLength;\n");
                    printf("            int rc = formatPDUonStream<STREAMTYPE>(outStream, pkt, reinterpret_cast<const PDUheader *>(nextPDU));\n");
                    printf("            return (rc);\n");
                    printf("          }\n");
                }
                printf("            break;\n");
            }
        }
        printf("        default:\n");
        printf("            // encapsulate in [{ .. }] to treat as parseable element\n");
        printf("            outStream << \"[UknownPDUtype={ \" << *pduHdr << \" }]\";\n");
        printf("            return (-1);\n");
        printf("        } // end switch on PDUtype\n");
        printf("    } // end pduHdr not null\n");
        printf("    return (0);\n");
        printf("}\n\n");
    }
}

/^enum/ {
    lastField=$NF;
    if (substr(lastField, length(lastField), 1) == "\r") {
        lastField = substr(lastField, 1, length(lastField) - 1);
        $NF = lastField; # replace with carriage return dropped
    }
    if ($1 == "enum") {
        group = $2;
        name = $3;
        val = $4;
        if (substr(val, 1, 1) == "(") { # strip surrounding parens
            val = substr(val, 2, length(val) - 2);
        }
        commentText = $5;
        for(i=6;length($i) > 0;i+=1) {
            commentText = commentText FS $i;
        }
        l = length(commentText);
        if (l > 2) {
            if (substr(commentText, 1, 1) == "\"") {
                if (substr(commentText, l, 1) == "\"") {
                    commentText = substr(commentText, 2, l - 2);
#printf("Stripped encapsulating quotes from %s\n", commentText) > "/dev/stderr";
                    l -= 2;
                } else {
#printf("Stripped leading quotes from %s\n", commentText) > "/dev/stderr";
                    commentText = substr(commentText, 2, l - 1);
                    l -= 1;
                }
            }
        }
        # tediously strip out any non-ASCII characters...
        cleanComment = "";
        for(i=1;i<=l;i+=1) {
            c = substr(commentText, i, 1);
            if (c <= "~") cleanComment = cleanComment c;
        }

        count = enumGroupCount[group] + 1;
        if (count == 1) { #
            lastEnumVal = -1;
        }
        enumGroupCount[group] = count;
        enumTableComment[group,count] = cleanComment;
        if (val != "") {
            enumTable[group,count] = name "=" val;
            lastEnumVal = val;
        } else {
                if (languageTarget != "python") {
                    enumTable[group,count] = name;
                } else {
                    lastEnumVal += 1;
                    enumTable[group,count] = name "=" lastEnumVal;
                }
        }
        next;
    }
}


{
    lastField=$NF;
    if (substr(lastField, length(lastField), 1) == "\r") {
        lastField = substr(lastField, 1, length(lastField) - 1);
        $NF = lastField; # replace with carriage return dropped
    }
#printf("NF=%d state=%d len(%s)=%d\n", NF, state, $1, length($1));
    if ((NF == 0) || (length($1) == 0) || ($1 == "END")) {
        if (state == 0) {
            next;   # skip empty lines
        }
        if (languageTarget == "python") {
            dumpPythonPDU_layout();
        } else {
            dumpPDU_wireLayout();
        }
        # add as new type to known table
#        renameType[PDUname] = "struct " PDUname;
#printf("Add %s size as %s%s\n", PDUname, currentOffset, dependentOffset);
        if (sawDependentField == 0) {
            sizeofTable[PDUname] = currentOffset dependentOffset;
        } else {
            variableLengthStructs[PDUname] = currentOffset dependentOffset;
        }
        # end PDU description
        if (languageTarget == "csharp") {
            dumpPDU_nativeLayout();
        }
        state = 0;
        next;
    }
    if (state == 0) {
        if (length($1) > 0) {
            # reset data 
            fieldCount = 0;
            currentOffset = 0;
            dependentOffset = "";
            lastDependentField = "";
            PDUname = $1;
printf("Start on PDU %s\n", PDUname) > "/dev/stderr";
            delete fieldNameToIndex;
            delete arrayElementCount;
            delete fieldDependsUpon;
            state = 1;
            next;
        }
        printf ("Searching for start of PDU description at line %d\n", NR) > "/dev/stderr";
        next;
    }
    if (state == 1) {
        if ($1 == "Name") {
            state = 2;
            next;
        }
        printf ("Searching for column heading at line %d\n", NR) > "/dev/stderr";
        next;
    }
    if (state == 2) {
        fN = stripSurroundingSpaces($1);
        fieldNameToIndex[fN] = fieldCount;
        fieldName[fieldCount] = fN;
        fieldType[fieldCount] = stripSurroundingSpaces($4);
        if (fieldType[fieldCount] == "") {
            printf("Error:  line %d has no type specified for %s\n", NR, fN) > "/dev/stderr";
            fieldType[fieldCount] = "void"; # ultimately, this won't compile
        }
        commentText = $5;
        if (lastDependentField != "") {
printf("Field %s depends on %s\n", fieldName[fieldCount], lastDependentField) > "/dev/stderr";
        }
        divideByMode = 0; # assume not variable length
        for(i=6;length($i) > 0;i+=1) {
            commentText = commentText FS $i;
        }
        l = length(commentText);
        if (l > 2) {
            if (substr(commentText, 1, 1) == "\"") {
                if (substr(commentText, l, 1) == "\"") {
                    commentText = substr(commentText, 2, l - 2);
#printf("Stripped encapsulating quotes from %s\n", commentText) > "/dev/stderr";
                    l -= 2;
                } else {
printf("Stripped leading quotes from %s\n", commentText) > "/dev/stderr";
                    commentText = substr(commentText, 2, l - 1);
                    l -= 1;
                }
            }
        }
        # tediously strip out any non-ASCII characters...
        cleanComment = "";
        for(i=1;i<=l;i+=1) {
            c = substr(commentText, i, 1);
            if (c <= "~") cleanComment = cleanComment c;
        }
        fieldComment[fieldCount] = cleanComment;
        i = index(fieldType[fieldCount], "[]");
        if (i == 0) { # not found
            i = index(fieldType[fieldCount], "[");
            if (i == 0) { # no subscripting at all
                baseType = fieldType[fieldCount];
                fieldMultiplier = 1;
                arrayElementCount[fieldCount] = 0;
                if (baseType in variableLengthStructs) {
                    fieldSize[fieldCount] = "get_" fN ".get_expectedSize()";
                    if (lastDependentField == "") {
                        lastDependentField = fieldName[fieldCount - 1];
                    }
                } else {
                    fieldSize[fieldCount] = sizeofTable[baseType];
                }
            } else { # numerical subscript
                baseType = substr(fieldType[fieldCount], 1, i - 1);
                j = index(fieldType[fieldCount], "]") - 1;
                fieldMultiplier = substr(fieldType[fieldCount],
                    i + 1, j - i);
                firstChar = substr(fieldMultiplier, 1, 1);
                if ((firstChar >= "0") && (firstChar <= "9")) {
                    arrayElementCount[fieldCount] = fieldMultiplier;
                } else {
printf("Variable length field %d %s size depends upon %s, prior offset=%d baseType=%s\n",
fieldCount, fieldName[fieldCount], fieldMultiplier, fieldOffset[fieldCount-1], baseType) > "/dev/stderr";
                    if (lastDependentField == "") {
                        lastDependentField = fieldName[fieldCount - 1];
                        dependentOffset = ""; # fieldOffset[fieldCount - 1];
#printf("set lastDependentField to %s offset=%s\n", lastDependentField, fieldOffset[fieldCount - 1]) ;
                    }
                    arrayElementCount[fieldCount] = fieldMultiplier;
                    if (baseType in variableLengthStructs) {
                        # dangerous -- really need to iterate over all elements
                        fieldSize[fieldCount] = "(" fieldMultiplier " * " fN ".get_expectedSize() )";
                    } else {
                        fieldSize[fieldCount] = "(" fieldMultiplier " * " sizeofTable[baseType] ")";
                    }
                }
            }
        } else { # empty subscript, length comes from Length column
            fieldMultiplier = $3;
#printf("fieldName=%s set fieldMultiplier to %s\n", fieldName[fieldCount], fieldMultiplier);
            arrayElementCount[fieldCount] = fieldMultiplier;
            if (fieldMultiplier == "pad8") {
#printf("special padding field\n");
                divideByMode = 2;
            } else if (fieldMultiplier == "pad8.4") {
                divideByMode = 3;
            } else {
                divideByMode = 1;
            }
            baseType = substr(fieldType[fieldCount], 1, i - 1);
        }
        if (! (baseType in sizeofTable)) {
            if (baseType in variableLengthStructs) {
                len = "get_" fN "().get_expectedSize()";
#printf("Variable length struct %s size is %s fieldSize=%s\n", baseType, len, fieldSize[fieldCount]);
            } else {
                printf ("Unknown type %s at line %d: %s\n",
                    baseType, NR, $0) > "/dev/stderr";
                len = 1;
            }
        } else {
            len = sizeofTable[baseType];
        }
        fieldBaseType[fieldCount] = baseType;
        fieldOffset[fieldCount] = sprintf("%d%s",
            currentOffset, dependentOffset);
        if (divideByMode == 1) { # field length comes from fieldMultiplier
            fieldLength[fieldCount] = fieldMultiplier / len;
            if (lastDependentField == "") {
                currentOffset += fieldMultiplier;
            }
            thisFieldSize = sprintf(" + (%s)", fieldMultiplier);    
#printf("fieldMultiplier=%s len=%d\n", fieldMultiplier, len);
        } else if (divideByMode == 2) { # special case mod function, pad to multiple of 8
            # padNeeded(x, w) = (w - (x & (w-1)) & (w-1))
            thisFieldSize = sprintf(" + ((8 - ((%s) & 7)) & 7)",
                fieldOffset[fieldCount]);
        } else if (divideByMode == 3) { # special case mod function, pad to multiple of 8+4
            # padNeeded(x, w) = (w - (x & (w-1)) & (w-1))
            thisFieldSize = sprintf(" + ((8 - (((%s) + 4) & 7)) & 7)",
                fieldOffset[fieldCount]);
        } else {
            fieldLength[fieldCount] = fieldMultiplier * len;
            if (lastDependentField == "") {
                currentOffset += fieldMultiplier * len;
            }
            if (fieldMultiplier == 1) {
                    thisFieldSize = " + " len;
            } else {
                if (len != 1) {
                    thisFieldSize = sprintf(" + (get_%s() * %s)",
                        fieldMultiplier, len);  
                } else {
                    thisFieldSize = sprintf(" + get_%s()",
                        fieldMultiplier);   
                }
            }
        }
        fieldDependsUpon[fieldCount] = lastDependentField;
        if (lastDependentField != "") { # start tracking dependency
            dependentOffset = dependentOffset thisFieldSize;
#printf("fieldName=%s fieldOffset=%s dependentOffset=%s\n", fieldName[fieldCount], fieldOffset[fieldCount], dependentOffset);
#           lastDependentField = fieldName[fieldCount];
        }
        fieldCount += 1;
    }
}

END {
    dumpEnumTables();
    if (languageTarget == "c++") {
        dumpEnumMacros();
        printf("\n#ifdef __CRT_WARN_NOT_ALREADY_DEFINED\n");
        printf("#undef __CRT_WARN_NOT_ALREADY_DEFINED\n");
        printf("#undef _CRT_SECURE_NO_WARNINGS /* restore original state */\n");
        printf("#endif\n\n");
        printf("#endif /* %s */\n", defineName);
    }
    printf("%s vim: set expandtab shiftwidth=4 tabstop=4:%s\n",
           startCommentChar, endCommentChar);
}

# /* vim: set expandtab shiftwidth=4 tabstop=4: */
