#ifndef _OME_EXEC_QUEUE_H
//! \cond
#define _OME_EXEC_QUEUE_H "$Id: OMEexecQueue.h 349 2020-05-21 04:19:56Z geoff $"
//! \endcond

#include <deque>

#include <OMEobject.h>
#include <OMEthread.h>

/** \defgroup OMEqueue OME Execution Queue
 * @{
 */

#define OME_MAX_PRIORITY_LEVELS	5
#define OME_DEFAULT_THREAD_PRIORITY 3

/*! \brief Implements thread queues with multiple levels
 * of priority.
 */
class OMEthreadExecQueue {
private:
	_STD deque<OMEthreadP>	threadList[OME_MAX_PRIORITY_LEVELS];
	uint_fast32_t   totalThreadsInQueue;
	uint_fast8_t    lastLevel; // up to OME_MAX_PRIORITY_LEVELS

	OME_FAST_CALL void finishSlice(OMEthread *);
public:
	OMEthreadExecQueue()	{
		totalThreadsInQueue = 0;
		lastLevel = 0;
	}

	~OMEthreadExecQueue()	{}

	OME_FAST_CALL void addThreadToQueue(OMEthread *t,
		int prio=OME_DEFAULT_THREAD_PRIORITY);

	OME_FAST_CALL bool removeSpecificThreadFromQueue(OMEthread *t, int prio=-1);

//	OME_FAST_CALL OMEthread *getNextThread(uint_fast8_t *prio = nullptr, bool restartAtTop=true);

	OME_FAST_CALL uint_fast16_t getNextThreads(OMEthread *v[],
		const uint_fast16_t maxWanted, uint_fast8_t *prio = nullptr, bool restartAtTop=true);

	OME_FAST_CALL int terminateThreadsOnObject(const OMEobject *obj);

	OME_FAST_CALL uint_fast32_t executeSlices(const uint_fast32_t maxCount);

	OME_FAST_CALL uint_fast32_t parallelExecuteSlices(const uint_fast32_t maxCount);

	uint_fast32_t totalQueuedAtLevel(uint_fast8_t l) const OME_ALWAYS_INLINE { return (threadList[l].size()); }

	uint_fast32_t totalQueued() const OME_ALWAYS_INLINE { return (totalThreadsInQueue); }

}; // end class OMEthreadExecQueue


// Global functions..., should these be static?
void OMEinitializeOIL2threadQueue(const int cpuTotal = 1, const int multiThreshold = 0);

int OMEexecuteOIL2threads(const uint_fast32_t maxToRun);

void OMEqueueOIL2thread(OMEthread *);

int OMEterminateThreadsOnObject(const OMEobject *obj);

uint_fast32_t OMEwaitForThreadsToStop(uint_fast32_t maxWaitTime);

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
