#ifndef _OME_MUTEX_H
//! \cond
#define _OME_MUTEX_H "$Id: OMEmutex.h 466 2020-07-24 23:00:54Z fargos $"
//! \endcond

/*! \file
 * \brief OME mutex and condition variable implementation.
 */

/** \addtogroup OMEutilFunctions
 * @{
 */

#    ifdef _WIN32
// WINDOWS
#    include <winsock2.h>
#    define DONT_USE_PTHREADS
#    define OME_MUTEX_TYPE CRITICAL_SECTION
#    define OME_CONDITION_TYPE HANDLE
// END WINDOWS
#    else
#    ifndef DONT_USE_PTHREADS
// Using POSIX Threads...
#        include <unistd.h>
#        include <pthread.h>
#        define OME_MUTEX_TYPE pthread_mutex_t
#        define OME_CONDITION_TYPE pthread_cond_t
#    else
#        define OME_MUTEX_TYPE uint32_t
#        define OME_CONDITION_TYPE uint32_t
// END POSIX THREADS
#    endif
#    endif /* if Windows or POSIX Threads */

#    include <OMEmanifests.h>

// Labels for Critical Sections within core.  Make sure
//  OME_CRITICAL_SECTION_TOTAL is updated to hold total count.
enum eOMEcriticalSectionLabel {
    OME_CRITICAL_SECTION_METHOD_INVOKE,
    OME_CRITICAL_SECTION_THREAD_QUEUE,
    OME_CRITICAL_SECTION_OID_REFERENCE,
    OME_CRITICAL_SECTION_EVENT_TABLE,
    OME_CRITICAL_SECTION_KERNEL_THREAD,
    OME_CRITICAL_SECTION_OMETYPE_OP,
    OME_CRITICAL_SECTION_IO_TABLE,
    OME_CRITICAL_SECTION_MISCELLANEOUS,
    OME_CRITICAL_SECTION_OMETYPE_REF,

    OME_CRITICAL_SECTION_TOTAL //!< count of critical sections...
};

OME_DLL_EXPORT void OMEstartCriticalSection(eOMEcriticalSectionLabel regionID);

OME_DLL_EXPORT void OMEendCriticalSection(eOMEcriticalSectionLabel regionID);

OME_DLL_EXPORT bool OMEinitStandardCriticalSectionMutexes();

OME_DLL_EXPORT bool OMEinitCriticalSectionMutex(OME_MUTEX_TYPE *, bool recursive = false) NONNULL_PARAMETERS(1);

OME_DLL_EXPORT int OMEdestroyCriticalSectionMutex(OME_MUTEX_TYPE *) NONNULL_PARAMETERS(1);

OME_DLL_EXPORT uint32_t OMEgetMiscellaneousMutex() OME_CONST_FUNCTION;

OME_DLL_EXPORT int OMEinitConditionVariable(OME_CONDITION_TYPE *);

OME_DLL_EXPORT int OMEdestroyConditionVariable(OME_CONDITION_TYPE *) NONNULL_PARAMETERS(1);

OME_DLL_EXPORT int OMEwaitForCondition(OME_CONDITION_TYPE *, eOMEcriticalSectionLabel mutexId) NONNULL_PARAMETERS(1);

OME_DLL_EXPORT int OMEsetConditionVariable(OME_CONDITION_TYPE *) NONNULL_PARAMETERS(1);

typedef void *(*OMEnativeTaskFP)(void *);

OME_DLL_EXPORT uint64_t OMEcallWithNativeThread(OMEnativeTaskFP func, void *functionArg);

OME_DLL_EXPORT int OMEwaitForNativeThread(uint64_t pid, void **retValue);

#define OME_MAX_CPUS_PERMITTED 128

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
