#ifndef _OME_SOCKET_H
//! \cond
#define _OME_SOCKET_H "$Id: OMEsocket.h 357 2020-06-18 23:13:25Z geoff $"
//! \endcond

#include <OMEioObjects.h>

/** \addtogroup OMEio
 * @{
 */

class OMEsocketDescriptor : public OMEioDescriptor {
protected:
static OMEioDescriptor *openSocket(struct sockaddr *sock, int sockLen,
	uint32_t openMode, int *errRet = nullptr);
public:
static OMEioDescriptor *openSocketScheme(const OMEstring &url, int *errRet = nullptr);
protected:
	uint32_t	    savedOpenMode;
	int			    openSides;
	bool			selectedForRead;
	bool			selectedForWrite;
#ifdef _WIN32
	static bool selectCallback(OMEevent *event, OMEthread *thread);
	WSAEVENT		eventHandle;
	uint32_t			eventMask;
	uint32_t			activeEvents;
	OMEeventCallbackFP	readCallback;
	OMEeventCallbackFP	writeCallback;

#endif
public:
	OMEsocketDescriptor(uint64_t handle, OMEioDescriptor::OMEopenMode t);

	~OMEsocketDescriptor();

#ifdef _WIN32
    // the WSAEVENT is the selectable handle, not the descriptor
    virtual uint64_t getSelectHandle() const override {
        return (reinterpret_cast<const uint64_t>(eventHandle));
    }
#endif

	virtual int readBytes(unsigned char *bfr, size_t bytes, size_t *actualBytes, int *errRet) override;

	virtual int writeBytes(const unsigned char *bfr, size_t bytes, size_t *actualBytes, int *errRet) override;

#ifndef _WIN32
	virtual int writeVectorOfBytes(const OMEioVector *v, uint_fast16_t elementCount,
		size_t *actualBytes, int *errRet, size_t skipBytes = 0) override;
#endif
	virtual OMEevent *selectForRead(OMEeventCallbackFP func, OMEthread *t = nullptr) override;

	virtual OMEevent *selectForWrite(OMEeventCallbackFP func, OMEthread *t = nullptr) override;

	virtual int close(int mask) override;

	virtual OMEstring *getLocalAddress() const override;

	virtual OMEstring *getPeerAddress() const override;

	virtual int getErrorStatus() const override;

	virtual bool isOpen() const override { return (openSides != 0); }

	virtual OMEioDescriptor *acceptConnection(int *errRet) override;

	virtual int receiveDatagram(unsigned char *bfr, size_t bytes,
            size_t *actualBytes, OMEstring *fromInfo, int *errRet) override;

	virtual int sendDatagram(const unsigned char *bfr, size_t bytes,
            size_t *actualBytes, const char *toInfo, int *errRet) override;
}; // end class OMEsocketDescriptor

class OMEunixDomainDescriptor : public OMEsocketDescriptor {
protected:
	OMEstring	socketFileName;
public:
	OMEunixDomainDescriptor(uint64_t handle, OMEioDescriptor::OMEopenMode t,
		const char *fileName);

	~OMEunixDomainDescriptor();
}; // end class OMEunixDomainDescriptor

void OMEregisterStandardSocketSchemes();

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
