#ifndef _CONCAT_HASH_H_
//! \cond
#define _CONCAT_HASH_H_ "$Id: concat_hash.h 454 2020-07-23 20:22:23Z geoff $"
//! \endcond

/* Copyright (C) 2010 - 2019 FARGOS Development, LLC */

/*! \file
 * \brief Concatenable hash algorithms
 *
 * Implementation of concatenable hash algorithms from:
 * http://oai.cwi.nl/oai/asset/14002/14002D.pdf
 * "Generic and Updatable XML Value Indices Covering Equality and Range Lookups"
 */


#include <stdint.h>
#include <utils/base/compiler_hints.h>

#ifdef __cplusplus
extern "C" {
#endif

/** \addtogroup FARGOSbase
 * @{
 */

/*! \addtogroup FARGOShash Concatenable Hash Functions
 * \brief Hash functions that support concatenation; logically:
 *  hash(A + B) = hash(a) + hash(b);
 * @{
 */

/*! \brief Generate a concatenateable 32-bit hash of a null-terminated string 
 * \param str points at the string to be hashed
 */
extern uint32_t hash_string32(const char *str);

/*! \brief Generate a concatenateable 32-bit hash of a block of memory
 *
 * \param block points at the start of the block to be hashed
 * \param blockLen specifies the length of the block
 */
extern uint32_t hash_block32(const void *block, size_t blockLen);

/*! \brief Compose a new hash of a logically concatenated string from
 * previous hashes of the two elements.
 *
 * \param left is the hash of the left-hand component
 * \param right is the hash of the right-hand component
 */
extern uint32_t hash_concat32(uint32_t left, uint32_t right) OME_CONST_FUNCTION;


/** @} */

/** @} */

#ifdef __cplusplus
}; /* close extern "C" scope */
#endif

#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
